/*
 * Jalview - A Sequence Alignment Editor and Viewer ($$Version-Rel$$)
 * Copyright (C) $$Year-Rel$$ The Jalview Authors
 * 
 * This file is part of Jalview.
 * 
 * Jalview is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *  
 * Jalview is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty 
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 * PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jalview.  If not, see <http://www.gnu.org/licenses/>.
 * The Jalview Authors are detailed in the 'AUTHORS' file.
 */
package jalview.analysis.scoremodels;

import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertTrue;

import jalview.datamodel.AlignmentI;
import jalview.datamodel.SequenceFeature;
import jalview.datamodel.SequenceI;
import jalview.gui.AlignFrame;
import jalview.gui.JvOptionPane;
import jalview.io.DataSourceType;
import jalview.io.FileLoader;
import jalview.math.MatrixI;

import java.util.Arrays;

import org.testng.Assert;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.Test;

public class FeatureDistanceModelTest
{

  @BeforeClass(alwaysRun = true)
  public void setUpJvOptionPane()
  {
    JvOptionPane.setInteractiveMode(false);
    JvOptionPane.setMockResponse(JvOptionPane.CANCEL_OPTION);
  }

  public static String alntestFile = "FER1_MESCR/72-76 DVYIL\nFER1_SPIOL/71-75 DVYIL\nFER3_RAPSA/21-25 DVYVL\nFER1_MAIZE/73-77 DVYIL\n";

  int[] sf1 = new int[] { 74, 74, 73, 73, 23, 23, -1, -1 };

  int[] sf2 = new int[] { -1, -1, 74, 75, -1, -1, 76, 77 };

  int[] sf3 = new int[] { -1, -1, -1, -1, -1, -1, 76, 77 };

  /**
   * <pre>
   * Load test alignment and add features to sequences: 
   *      FER1_MESCR FER1_SPIOL FER3_RAPSA FER1_MAIZE 
   *  sf1     X          X          X  
   *  sf2                X                     X 
   *  sf3                                      X
   * </pre>
   * 
   * @return
   */
  public AlignFrame getTestAlignmentFrame()
  {
    AlignFrame alf = new FileLoader(false).LoadFileWaitTillLoaded(
            alntestFile, DataSourceType.PASTE);
    AlignmentI al = alf.getViewport().getAlignment();
    Assert.assertEquals(al.getHeight(), 4);
    Assert.assertEquals(al.getWidth(), 5);
    for (int i = 0; i < 4; i++)
    {
      SequenceI ds = al.getSequenceAt(i).getDatasetSequence();
      if (sf1[i * 2] > 0)
      {
        ds.addSequenceFeature(new SequenceFeature("sf1", "sf1", "sf1",
                sf1[i * 2], sf1[i * 2 + 1], "sf1"));
      }
      if (sf2[i * 2] > 0)
      {
        ds.addSequenceFeature(new SequenceFeature("sf2", "sf2", "sf2",
                sf2[i * 2], sf2[i * 2 + 1], "sf2"));
      }
      if (sf3[i * 2] > 0)
      {
        ds.addSequenceFeature(new SequenceFeature("sf3", "sf3", "sf3",
                sf3[i * 2], sf3[i * 2 + 1], "sf3"));
      }
    }
    alf.setShowSeqFeatures(true);
    alf.getFeatureRenderer().setVisible("sf1");
    alf.getFeatureRenderer().setVisible("sf2");
    alf.getFeatureRenderer().setVisible("sf3");
    alf.getFeatureRenderer().findAllFeatures(true);
    Assert.assertEquals(alf.getFeatureRenderer().getDisplayedFeatureTypes()
            .size(), 3, "Number of feature types");
    assertTrue(alf.getCurrentView().areFeaturesDisplayed());
    return alf;
  }

  @Test(groups = { "Functional" })
  public void testFeatureScoreModel() throws Exception
  {
    AlignFrame alf = getTestAlignmentFrame();
    FeatureDistanceModel fsm = new FeatureDistanceModel();
    assertTrue(fsm.configureFromAlignmentView(alf.getCurrentView()
            .getAlignPanel()));
    alf.selectAllSequenceMenuItem_actionPerformed(null);

    MatrixI dm = fsm
            .findDistances(alf.getViewport().getAlignmentView(
            true));
    assertEquals(dm.getValue(0, 2), 0d,
            "FER1_MESCR (0) should be identical with RAPSA (2)");
    assertTrue(dm.getValue(0, 1) > dm.getValue(0, 2),
            "FER1_MESCR (0) should be further from SPIOL (1) than it is from RAPSA (2)");
  }

  @Test(groups = { "Functional" })
  public void testFeatureScoreModel_hiddenFirstColumn() throws Exception
  {
    AlignFrame alf = getTestAlignmentFrame();
    // hiding first two columns shouldn't affect the tree
    alf.getViewport().hideColumns(0, 1);
    FeatureDistanceModel fsm = new FeatureDistanceModel();
    assertTrue(fsm.configureFromAlignmentView(alf.getCurrentView()
            .getAlignPanel()));
    alf.selectAllSequenceMenuItem_actionPerformed(null);
    MatrixI dm = fsm
            .findDistances(alf.getViewport().getAlignmentView(
            true));
    assertEquals(dm.getValue(0, 2), 0d,
            "FER1_MESCR (0) should be identical with RAPSA (2)");
    assertTrue(dm.getValue(0, 1) > dm.getValue(0, 2),
            "FER1_MESCR (0) should be further from SPIOL (1) than it is from RAPSA (2)");
  }

  @Test(groups = { "Functional" })
  public void testFeatureScoreModel_HiddenColumns() throws Exception
  {
    AlignFrame alf = getTestAlignmentFrame();
    // hide columns and check tree changes
    alf.getViewport().hideColumns(3, 4);
    alf.getViewport().hideColumns(0, 1);
    FeatureDistanceModel fsm = new FeatureDistanceModel();
    assertTrue(fsm.configureFromAlignmentView(alf.getCurrentView()
            .getAlignPanel()));
    alf.selectAllSequenceMenuItem_actionPerformed(null);
    MatrixI dm = fsm
            .findDistances(alf.getViewport().getAlignmentView(
            true));
    assertEquals(
            dm.getValue(0, 2),
            0d,
            "After hiding last two columns FER1_MESCR (0) should still be identical with RAPSA (2)");
    assertEquals(
            dm.getValue(0, 1),
            0d,
            "After hiding last two columns FER1_MESCR (0) should now also be identical with SPIOL (1)");
    for (int s = 0; s < 3; s++)
    {
      assertTrue(dm.getValue(s, 3) > 0d, "After hiding last two columns "
              + alf.getViewport().getAlignment().getSequenceAt(s).getName()
              + "(" + s + ") should still be distinct from FER1_MAIZE (3)");
    }
  }

  /**
   * Check findFeatureAt doesn't return contact features except at contact
   * points TODO:move to under the FeatureRendererModel test suite
   */
  @Test(groups = { "Functional" })
  public void testFindFeatureAt_PointFeature() throws Exception
  {
    String alignment = "a CCCCCCGGGGGGCCCCCC\n" + "b CCCCCCGGGGGGCCCCCC\n"
            + "c CCCCCCGGGGGGCCCCCC\n";
    AlignFrame af = new jalview.io.FileLoader(false)
            .LoadFileWaitTillLoaded(alignment, DataSourceType.PASTE);
    SequenceI aseq = af.getViewport().getAlignment().getSequenceAt(0);
    SequenceFeature sf = null;
    sf = new SequenceFeature("disulphide bond", "", 2, 5, Float.NaN, "");
    aseq.addSequenceFeature(sf);
    assertTrue(sf.isContactFeature());
    af.refreshFeatureUI(true);
    af.getFeatureRenderer().setAllVisible(Arrays.asList("disulphide bond"));
    Assert.assertEquals(af.getFeatureRenderer().getDisplayedFeatureTypes()
            .size(), 1, "Should be just one feature type displayed");
    // step through and check for pointwise feature presence/absence
    Assert.assertEquals(af.getFeatureRenderer().findFeaturesAtRes(aseq, 1)
            .size(), 0);
    // step through and check for pointwise feature presence/absence
    Assert.assertEquals(af.getFeatureRenderer().findFeaturesAtRes(aseq, 2)
            .size(), 1);
    // step through and check for pointwise feature presence/absence
    Assert.assertEquals(af.getFeatureRenderer().findFeaturesAtRes(aseq, 3)
            .size(), 0);
    // step through and check for pointwise feature presence/absence
    Assert.assertEquals(af.getFeatureRenderer().findFeaturesAtRes(aseq, 4)
            .size(), 0);
    // step through and check for pointwise feature presence/absence
    Assert.assertEquals(af.getFeatureRenderer().findFeaturesAtRes(aseq, 5)
            .size(), 1);
    // step through and check for pointwise feature presence/absence
    Assert.assertEquals(af.getFeatureRenderer().findFeaturesAtRes(aseq, 6)
            .size(), 0);
  }

  @Test(groups = { "Functional" })
  public void testFindDistances() throws Exception
  {
    String seqs = ">s1\nABCDE\n>seq2\nABCDE\n";
    AlignFrame alf = new FileLoader().LoadFileWaitTillLoaded(seqs,
            DataSourceType.PASTE);
    SequenceI s1 = alf.getViewport().getAlignment().getSequenceAt(0);
    SequenceI s2 = alf.getViewport().getAlignment().getSequenceAt(1);

    /*
     * set domain and variant features thus:
     *     ----5
     *  s1 ddd..
     *  s1 .vvv.
     *  s1 ..vvv    
     *  s2 .ddd. 
     *  s2 vv..v
     *  The number of unshared feature types per column is
     *     20120 (two features of the same type doesn't affect score)
     *  giving an average (pairwise distance) of 5/5 or 1.0 
     */
    s1.addSequenceFeature(new SequenceFeature("domain", null, 1, 3, 0f,
            null));
    s1.addSequenceFeature(new SequenceFeature("variant", null, 2, 4, 0f,
            null));
    s1.addSequenceFeature(new SequenceFeature("variant", null, 3, 5, 0f,
            null));
    s2.addSequenceFeature(new SequenceFeature("domain", null, 2, 4, 0f,
            null));
    s2.addSequenceFeature(new SequenceFeature("variant", null, 1, 2, 0f,
            null));
    s2.addSequenceFeature(new SequenceFeature("variant", null, 5, 5, 0f,
            null));
    alf.setShowSeqFeatures(true);
    alf.getFeatureRenderer().findAllFeatures(true);

    FeatureDistanceModel fsm = new FeatureDistanceModel();
    assertTrue(fsm.configureFromAlignmentView(alf.getCurrentView()
            .getAlignPanel()));
    alf.selectAllSequenceMenuItem_actionPerformed(null);

    MatrixI distances = fsm.findDistances(alf.getViewport()
            .getAlignmentView(true));
    assertEquals(distances.width(), 2);
    assertEquals(distances.height(), 2);
    assertEquals(distances.getValue(0, 0), 0d);
    assertEquals(distances.getValue(1, 1), 0d);
    // these left to fail pending resolution of
    // JAL-2424 (dividing score by 6, not 5)
    assertEquals(distances.getValue(0, 1), 1f);
    assertEquals(distances.getValue(1, 0), 1f);
  }

}
